/*
math.h - basic math operations
Copyright (C) 2024 Uncle Mike

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
*/

#pragma once

#include <assert.h>
#include <float.h>
#include <math.h>

// angle representation
#define PITCH	0
#define YAW	1
#define ROLL	2

#ifndef M_PI
#define M_PI			3.1415926535897932384626433832795f
#endif

// radian to degrees
#define RAD2DEG( x )		((x) * (180.0f / M_PI))
#define DEG2RAD( x )		((x) * (M_PI / 180.0f))

inline void SinCos( float angle, float &s, float &c ) 
{
	s = sinf( angle );
	c = cosf( angle );
}

struct vec2
{
	vec2( const vec2 &a ) { x = a.x; y = a.y; }
	vec2( const float *a ) { x = a[0]; y = a[1]; }
	vec2( float a0, float a1 ) { x = a0; y = a1; }
	vec2( float a ) { x = y = a; }
	vec2() {}

	vec2 &operator=( const vec2 &a ) { x = a.x; y = a.y; return *this; }
	vec2 operator-() const { return vec2(-x,-y); }
	vec2 &operator*=(float a) { x *= a; y *= a; return *this; }
	vec2 &operator/=(float a) { x /= a; y /= a; return *this; }
	vec2 &operator*=(const vec2 &a) { x *= a.x; y *= a.y; return *this; }
	vec2 &operator/=(const vec2 &a) { x /= a.x; y /= a.y; return *this; }
	vec2 &operator+=(const vec2 &a) { x += a.x; y += a.y; return *this; }
	vec2 &operator-=(const vec2 &a) { x -= a.x; y -= a.y; return *this; }
	operator float*() { return &x; }
	operator const float*() const { return &x; }

	float	x, y;
};

inline vec2 operator * ( const vec2 &a, const float b ) { return vec2( a.x * b, a.y * b ); }
inline vec2 operator * ( const float a, const vec2 &b ) { return vec2( a * b.x, a * b.y ); }
inline vec2 operator * ( const vec2 &a, const vec2 &b ) { return vec2( a.x * b.x, a.y * b.y ); }
inline vec2 operator / ( const vec2 &a, const float b ) { return vec2( a.x / b, a.y / b ); }
inline vec2 operator / ( const vec2 &a, const vec2 &b ) { return vec2( a.x / b.x, a.y / b.y ); }
inline vec2 operator / ( const float a, const vec2 &b ) { return vec2( a / b.x, a / b.y ); }
inline vec2 operator + ( const vec2 &a, const vec2 &b ) { return vec2( a.x + b.x, a.y + b.y ); }
inline vec2 operator + ( const vec2 &a, const float b ) { return vec2( a.x + b, a.y + b ); }
inline vec2 operator - ( const vec2 &a, const vec2 &b ) { return vec2( a.x - b.x, a.y - b.y ); }
inline vec2 operator - ( const vec2 &a, const float b ) { return vec2( a.x - b, a.y - b ); }

struct vec3
{
	vec3( const vec3& a ) { x = a.x; y = a.y; z = a.z; }
	vec3( const float *a ) { x = a[0]; y = a[1]; z = a[2]; }
	vec3( float a0, float a1, float a2 ) { x = a0; y = a1; z = a2; }
	vec3( float a ) { x = y = z = a; }
	vec3() {}

	vec3 &operator=(const vec3 &a) { x = a.x; y = a.y; z = a.z; return *this; }
	vec3 operator-() const { return vec3(-x,-y,-z); }
	vec3 &operator*=(float a) { x *= a; y *= a; z *= a; return *this; }
	vec3 &operator/=(float a) { x /= a; y /= a; z /= a; return *this; }
	vec3 &operator+=(const vec3 &a) { x += a.x; y += a.y; z += a.z; return *this; }
	vec3 &operator-=(const vec3 &a) { x -= a.x; y -= a.y; z -= a.z; return *this; }
	operator float*() { return &x; }
	operator const float*() const { return &x; }

	float	x, y, z;
};

inline vec3 operator * ( const vec3 &a, const float b ) { return vec3( a.x * b, a.y * b, a.z * b ); }
inline vec3 operator * ( const float a, const vec3 &b ) { return vec3( a * b.x, a * b.y, a * b.z ); }
inline vec3 operator * ( const vec3 &a, const vec3 &b ) { return vec3 (a.x * b.x, a.y * b.y, a.z * b.z); }
inline vec3 operator / ( const vec3 &a, const float b ) { return vec3( a.x / b, a.y / b, a.z / b ); }
inline vec3 operator / ( const float a, const vec3 &b ) { return vec3( a / b.x, a / b.y, a / b.z ); }
inline vec3 operator + ( const vec3 &a, const vec3 &b ) { return vec3( a.x + b.x, a.y + b.y, a.z + b.z ); }
inline vec3 operator + ( const vec3 &a, const float b ) { return vec3( a.x + b, a.y + b, a.z + b ); }
inline vec3 operator - ( const vec3 &a, const vec3 &b ) { return vec3( a.x - b.x, a.y - b.y, a.z - b.z ); }
inline vec3 operator - ( const vec3 &a, const float b ) { return vec3( a.x - b, a.y - b, a.z - b ); }

struct rad3
{
	rad3( const rad3& a ) { x = a.x; y = a.y; z = a.z; }
	rad3( float a0, float a1, float a2 ) { x = a0; y = a1; z = a2; }
	rad3( float a ) { x = y = z = a; }
	rad3() {}

	rad3 &operator=( const rad3 &a ) { x = a.x; y = a.y; z = a.z; return *this; }
	operator float*() { return &x; }
	operator const float*() const { return &x; }

	float	x, y, z;
};

struct vec4
{
	vec4(const vec2 &a) { x = a.x; y = a.y; z = 0.0; w = 0.0; }
	vec4(const vec3 &a) { x = a.x; y = a.y; z = a.z; w = 1.0; }
	vec4(const vec4 &a) { x = a.x; y = a.y; z = a.z; w = a.w; }
	vec4(const vec3 &a, float b) { x = a.x; y = a.y; z = a.z; w = b; }
	vec4(const float *a) { x = a[0]; y = a[1]; z = a[2]; w = a[3]; }
	vec4(float a0, float a1, float a2, float a3) { x = a0; y = a1; z = a2; w = a3; }
	vec4(float a) { x = y = z = w = a; }
	vec4() {}

	vec4 &operator=(const vec4 &a) { x = a.x; y = a.y; z = a.z; w = a.w; return *this; }
	vec4 operator-() const { return vec4(-x,-y,-z,-w); }
	vec4 &operator*=(float a) { x *= a; y *= a; z *= a; w *= a; return *this; }
	vec4 &operator/=(float a) { x /= a; y /= a; z /= a; w /= a; return *this; }
	vec4 &operator+=(const vec4 &a) { x += a.x; y += a.y; z += a.z; w += a.w; return *this; }
	vec4 &operator-=(const vec4 &a) { x -= a.x; y -= a.y; z -= a.z; w -= a.w; return *this; }
	operator float*() { return &x; }
	operator const float*() const { return &x; }

	float	x, y, z, w;
};

inline vec4 operator * ( const vec4 &a, const float b ) { return vec4( a.x * b, a.y * b, a.z * b, a.w * b ); }
inline vec4 operator * ( const float a, const vec4 &b ) { return vec4( a * b.x, a * b.y, a * b.z, a * b.w ); }
inline vec4 operator * ( const vec4 &a, const vec4 &b ) { return vec4( a.x * b.x, a.y * b.y, a.z * b.z, a.w * b.w ); }
inline vec4 operator / ( const vec4 &a, const float b ) { return vec4( a.x / b, a.y / b, a.z / b, a.w / b ); }
inline vec4 operator + ( const vec4 &a, const vec4 &b ) { return vec4( a.x + b.x, a.y + b.y, a.z + b.z, a.w + b.w ); }
inline vec4 operator + ( const vec4 &a, const float b ) { return vec4( a.x + b, a.y + b, a.z + b, a.w + b ); }
inline vec4 operator - ( const vec4 &a, const vec4 &b ) { return vec4( a.x - b.x, a.y - b.y, a.z - b.z, a.w - b.w ); }
inline vec4 operator - ( const vec4 &a, const float b ) { return vec4( a.x - b, a.y - b, a.z - b, a.w - b ); }

struct xform 
{
	xform() {}
	xform( const xform &other ) { mat[0] = other.mat[0]; mat[1] = other.mat[1]; mat[2] = other.mat[2]; mat[3] = other.mat[3]; }

	xform( const vec3 &origin, const vec3 &angles, const vec3 &scale = 1.0f )
	{
		float	sr, sp, sy, cr, cp, cy;

		SinCos( DEG2RAD( angles[YAW] ), sy, cy );
		SinCos( DEG2RAD( angles[PITCH] ), sp, cp );
		SinCos( DEG2RAD( angles[ROLL] ), sr, cr );

		mat[0].x = cp * cy * scale.x;
		mat[0].y = cp * sy * scale.x;
		mat[0].z = -sp * scale.x;
		mat[1].x = (sr * sp * cy + cr * -sy) * scale.y;
		mat[1].y = (sr * sp * sy + cr * cy) * scale.y;
		mat[1].z = sr * cp * scale.y;
		mat[2].x = (cr * sp * cy + -sr * -sy) * scale.z;
		mat[2].y = (cr * sp * sy + -sr * cy) * scale.z;
		mat[2].z = cr * cp * scale.z;
		mat[3].x = origin.x;
		mat[3].y = origin.y;
		mat[3].z = origin.z;
	}

	xform( const vec3 &origin, const rad3 &angles, const vec3 &scale = 1.0f )
	{
		float	sr, sp, sy, cr, cp, cy;

		SinCos( angles[ROLL], sy, cy );
		SinCos( angles[YAW], sp, cp );
		SinCos( angles[PITCH], sr, cr );

		mat[0].x = cp * cy * scale.x;
		mat[0].y = cp * sy * scale.x;
		mat[0].z = -sp * scale.x;
		mat[1].x = (sr * sp * cy + cr * -sy) * scale.y;
		mat[1].y = (sr * sp * sy + cr * cy) * scale.y;
		mat[1].z = sr * cp * scale.y;
		mat[2].x = (cr * sp * cy + -sr * -sy) * scale.z;
		mat[2].y = (cr * sp * sy + -sr * cy) * scale.z;
		mat[2].z = cr * cp * scale.z;
		mat[3].x = origin.x;
		mat[3].y = origin.y;
		mat[3].z = origin.z;
	}

	const bool identity( void ) const
	{
		if( mat[0].x != 1 || mat[0].y != 0 || mat[0].z != 0 )
			return false;
		if( mat[1].x != 0 || mat[1].y != 1 || mat[1].z != 0 )
			return false;
		if( mat[2].x != 0 || mat[2].y != 0 || mat[2].z != 1 )
			return false;
		if( mat[3].x != 0 || mat[3].y != 0 || mat[3].z != 0 )
			return false;
		return true;
	}

	void load_identity( void )
	{
		mat[0] = vec3( 1, 0, 0 );
		mat[1] = vec3( 0, 1, 0 );
		mat[2] = vec3( 0, 0, 1 );
		mat[3] = vec3( 0, 0, 0 );
	}

	xform &operator=(const xform &a) { mat[0] = a[0]; mat[1] = a[1]; mat[2] = a[2]; mat[3] = a[3]; return *this; }
	vec3 &operator[]( int i ) { assert( i >= 0 && i < 4 ); return mat[i]; }
	const vec3  &operator[]( int i ) const { assert( i >= 0 && i < 4 ); return mat[i]; }

	operator vec3() const
	{
		float	xyDist = sqrtf( mat[0].x * mat[0].x + mat[0].y * mat[0].y );
		vec3 	angles;

		// enough here to get angles?
		if( xyDist > FLT_EPSILON )
		{
			angles.x = RAD2DEG( atan2f( -mat[0].z, xyDist ));
			angles.y = RAD2DEG( atan2f( mat[0].y, mat[0].x ));
			angles.z = RAD2DEG( atan2f( mat[1].z, mat[2].z ));
		}
		else	// forward is mostly Z, gimbal lock
		{
			angles.x = RAD2DEG( atan2f( -mat[0].z, xyDist ));
			angles.y = RAD2DEG( atan2f( -mat[1].x, mat[1].y ));
			angles.z = 0;
		}

		return angles;
	}

	operator rad3() const
	{
		float	xyDist = sqrtf( mat[0].x * mat[0].x + mat[0].y * mat[0].y );
		rad3	angles;

		if( xyDist > FLT_EPSILON )
		{
			angles.y = atan2f( -mat[0].z, xyDist );
			angles.z = atan2f( mat[0].y, mat[0].x );
			angles.x = atan2f( mat[1].z, mat[2].z );
		}
		else	// forward is mostly Z, gimbal lock
		{
			angles.y = atan2f( -mat[0].z, xyDist );
			angles.z = atan2f( -mat[1].x, mat[1].y );
			angles.x = 0.0;
		}

		return angles;
	}

	vec3	mat[4];
};

// vec3 rotate
inline vec3 operator * (const xform &a, const vec3 &b)
{
	vec3	out;

	out.x = b.x * a[0].x + b.y * a[1].x + b.z * a[2].x;
	out.y = b.x * a[0].y + b.y * a[1].y + b.z * a[2].y;
	out.z = b.x * a[0].z + b.y * a[1].z + b.z * a[2].z;

	return out;
}

// vec3 irotate
inline vec3 operator * (const vec3 &a, const xform &b)
{
	vec3	out;

	out.x = a.x * b[0].x + a.y * b[0].y + a.z * b[0].z;
	out.y = a.x * b[1].x + a.y * b[1].y + a.z * b[1].z;
	out.z = a.x * b[2].x + a.y * b[2].y + a.z * b[2].z;

	return out;
}
