/*
model_csm.h - Compiled Static Mesh (CSM) model format description
Copyright (C) 2024 Uncle Mike

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
*/

#pragma once

#include "math.h"
#include "string_tools.h"
#include <vector>

// bit routines
#define BIT( n )			(1U<<( n ))
#define SetBits( iBitVector, bits )	((iBitVector) = (iBitVector) | (bits))
#define ClearBits( iBitVector, bits )	((iBitVector) = (iBitVector) & ~(bits))
#define FBitSet( iBitVector, bit )	((iBitVector) & (bit))

// built-in material names
#define MATERIAL_NULL		"@null"
#define MATERIAL_SKIP		"@skip"

// enumerate coordinate channels
#define FCHAN_TC			0
#define FCHAN_LM			1

#define FMODEL_LMGROUPS_COMPUTED	BIT( 0 )
#define FMODEL_BBOX_COMPUTED		BIT( 1 )
#define FMODEL_INDEXED_GEOMETRY	BIT( 2 )
#define FMODEL_KEEP_NORMALS		BIT( 3 )
#define FMODEL_DTGROUPS_COMPUTED	BIT( 4 )
#define FMODEL_WINDINGS_COMPUTED	BIT( 5 )
#define FMODEL_HAS_SOURCE_LMGROUPS	BIT( 6 )

#define FFACE_FROM_PLANAR_GROUP	BIT( 0 )
#define FFACE_FROM_PURE_AXIAL_GROUP	BIT( 1 )
#define FFACE_DISSOLVE_FIRST_EDGE	BIT( 2 )
#define FFACE_DISSOLVE_SECOND_EDGE	BIT( 3 )
#define FFACE_DISSOLVE_THIRD_EDGE	BIT( 4 )
#define FFACE_HAS_SOURCE_LMCOORDS	BIT( 5 )
#define FFACE_LANDSCAPE		BIT( 6 )
#define FFACE_STRUCTURAL		BIT( 7 )	// can be used for BSP-processing

#define FFACE_IS_CHECKED		BIT( 15 )	// internal flag
#define FFACE_WORLD_TARGET		(FFACE_FROM_PLANAR_GROUP|FFACE_FROM_PURE_AXIAL_GROUP)

/*
========================================================================

.CSM triangle model file format

========================================================================
*/

#define IDCSMMODHEADER		(('M'<<24)+('S'<<16)+('C'<<8)+'I') // little-endian "ICSM"
#define IDCSM_VERSION		3
#define CSM_MAXNAME			260

// abstract header for catch changes
struct csm_header_base_t
{
	unsigned int	ident;
	unsigned int	version;
	unsigned int	header_size;	// sizeof( csm_header_t )
	unsigned int	flags;
};

struct csm_header_t
{
	unsigned int	ident;
	unsigned int	version;
	unsigned int	header_size;	// sizeof( csm_header_t )
	unsigned int	flags;

	char		pathes[1024];	// relative paths to textures (used for modelviewers)
	unsigned int	lmGroups;		// light group count
	unsigned int	dtGroups;		// detail group count
	vec3		model_mins;
	vec3		model_maxs;

	// CSM_VERSION == 2
	unsigned int	mat_ofs;
	unsigned int	mat_size;

	unsigned int	faces_ofs;
	unsigned int	face_size;	// sizeof( csm_face_t )
	unsigned int	faces_count;

	unsigned int	vertex_ofs;
	unsigned int	vertex_size;	// sizeof( csm_vertex_t )
	unsigned int	vertex_count;

	// CSM_VERSION == 3
	unsigned int	sides_ofs;
	unsigned int	side_size;	// sizeof( csm_side_t )
	unsigned int	sides_count;

	unsigned int	points_ofs;
	unsigned int	point_size;	// sizeof( vec3 )
	unsigned int	points_count;
};

struct csm_color_t
{
	csm_color_t() { r = g = b = a = 255; }

	unsigned char	r, g, b, a;
};

// CSM model vertex
struct csm_vertex_t
{
	csm_vertex_t() : point(0.0f), normal(0.0f) {}
	csm_vertex_t( const vec3 &p ) : point(p), normal(0.0f) {}
	csm_vertex_t( const vec3 &p, const vec3 &n ) : point(p), normal(n) {}

	vec3		point;
	vec3		normal;
	csm_color_t	color;
};

struct csm_uv_t
{
	csm_uv_t() { uv[0] = 0.0f, uv[1] = 0.0f, uv[2] = 0.0f; }

	vec2		uv[3];
};

// CSM model face
struct csm_face_t
{
	csm_face_t() : material(0), flags(0), lmGroup(-1), dtGroup(-1) { memset( index, 0, sizeof( index )); }

	unsigned short	material;
	unsigned short	flags;
	unsigned int	index[3];
	int		lmGroup;
	int		dtGroup;
	csm_uv_t		tc[2];
};

// CSM model poly
struct csm_side_t
{
	unsigned int	firstpoint;
	unsigned short	numpoints;
	unsigned short	material;
	int		brushGroup;	// -1 if single winding
	unsigned int	reserved;		// for future expansions
	vec4		vecs[2];		// texture coordinate mapping

	csm_side_t() : material(0), firstpoint(0), numpoints(0) {}
};

/*
========================================================================

intermediate model struct

========================================================================
*/
struct material_t
{
	char		name[CSM_MAXNAME];
};

struct model_t
{
	model_t() : file(0), flags(0), lmGroups(0), mins(FLT_MAX), maxs(-FLT_MAX) {}

	std::vector<material_t>	materials;
	std::vector<csm_vertex_t>	verts;
	std::vector<csm_face_t>	faces;
	unsigned int		flags;
	unsigned int		lmGroups;	// total count of lightmap groups
	unsigned int		dtGroups;
	vec3			mins, maxs;
	FILE			*file;

	const char *material( const csm_face_t *F ) const
	{
		assert( F != NULL );
		assert( F->material >= 0 && F->material < materials.size() );
		return materials[F->material].name;
	}

	unsigned short AllocMaterial( const char *materialname )
	{
		material_t	newMat;

		for( size_t i = 0; i < materials.size(); i++ )
		{
			if( !stricmp( materials[i].name, materialname ))
				return i;
		}

		strncpy( newMat.name, materialname, sizeof( newMat.name ) - 1 );
		materials.push_back( newMat );

		return materials.size() - 1;
	}

	bool ParseMaterial( const unsigned char *&data )
	{
		char	token[CSM_MAXNAME];
		int	c, len;

		token[0] = 0;
		len = 0;
	
		if( !data )
			return false;

		// skip whitespace
		while(( c = *data ) <= ' ' )
		{
			if( c == 0 )
				return false; // end of file;
			data++;
		}

		if( c == '\"' )
		{
			data++;
			while( 1 )
			{
				c = *data++;
				// handle strings like """
				if(( c == '\"' && *data != '\"' ) || !c )
				{
					if( len < CSM_MAXNAME )
						token[len] = 0;
					// alloc material name
					AllocMaterial( token );
					return true;
				}

				if( len < CSM_MAXNAME )
					token[len] = c;
				len++;
			}
		}

		// all material tokens should be isolated with double quoutes
		return false;
	}

	bool load( const char *name )
	{
		unsigned char	*mat_buf;
		unsigned char	*mat_ptr;
		csm_header_base_t	hdr_base;
		csm_header_t	header;
		csm_face_t	face;
		csm_vertex_t	vert;
		int		i;

		if(( file = fopen( name, "rb" )) == NULL )
		{
			printf( "failed to read: %s\n", name );
			return false;
		}

		if( fread( &hdr_base, 1, sizeof( hdr_base ), file ) != sizeof( hdr_base ))
		{
			printf( "%s: unexcpected end of file\n", name );
			fclose( file );
			file = NULL;
			return false;
		}

		if( hdr_base.ident != IDCSMMODHEADER )
		{
			printf( "%s: has wrong id (%x should be %x)\n", name, hdr_base.ident, IDCSMMODHEADER );
			fclose( file );
			file = NULL;
			return false;
		}

		if( hdr_base.version != IDCSM_VERSION )
		{
			printf( "%s: has wrong version number (%i should be %i)\n", name, hdr_base.version, IDCSM_VERSION );
			fclose( file );
			file = NULL;
			return false;
		}

		if( hdr_base.header_size < sizeof( csm_header_t ))
		{
			printf( "%s too small header (%i should be %i)\n", name, hdr_base.header_size, sizeof( csm_header_t ));
			fclose( file );
			file = NULL;
			return false;
		}

		// loader can't read a new features, it's will be just ignored
		if( hdr_base.header_size > sizeof( csm_header_t ))
			hdr_base.header_size = sizeof( csm_header_t );

		// rewind to start
		fseek( file, 0, SEEK_SET );

		// read a real header
		if( fread( &header, 1, hdr_base.header_size, file ) != hdr_base.header_size )
		{
			printf( "%s: unexcpected end of file\n", name );
			fclose( file );
			file = NULL;
			return false;
		}

		if( !header.faces_count )
		{
			printf( "%s: has no faces\n", name );
			fclose( file );
			file = NULL;
			return false;
		}

		if( header.face_size != sizeof( csm_face_t ))
		{
			printf( "%s: mismatch face size (%i should be %i)\n", name, header.face_size, sizeof( csm_face_t ));
			fclose( file );
			file = NULL;
			return false;
		}

		if( !header.vertex_count )
		{
			printf( "%s: has no verts\n", name );
			fclose( file );
			file = NULL;
			return false;
		}

		if( header.vertex_size != sizeof( csm_vertex_t ))
		{
			printf( "%s: mismatch vert size (%i should be %i)\n", name, header.vertex_size, sizeof( csm_vertex_t ));
			fclose( file );
			file = NULL;
			return false;
		}

		// parse materials
		// historically CSM doesn't have stringtable only material names
		fseek( file, header.mat_ofs, SEEK_SET );
		mat_buf = (unsigned char *)malloc( header.mat_size + 1 );
		memset( mat_buf, 0, header.mat_size + 1 );

		if( fread( mat_buf, 1, sizeof( unsigned char ) * header.mat_size, file ) != ( sizeof( unsigned char ) * header.mat_size ))
		{
			printf( "%s: unexpected end of file while loading materials\n", name );
			fclose( file );
			file = NULL;
			return false;
		}

		// all tests are passed
		// now loading the model
		materials.clear();
		faces.clear();
		verts.clear();

		flags = header.flags;
		lmGroups = header.lmGroups;
		dtGroups = header.dtGroups;
		mins = header.model_mins;
		maxs = header.model_maxs;

		mat_ptr = mat_buf;
		while( ParseMaterial( mat_ptr ));
		free( mat_buf );

		// load faces
		fseek( file, header.faces_ofs, SEEK_SET );
		for( i = 0; i < header.faces_count; i++ )
		{
			if( fread( &face, 1, sizeof( face ), file ) != sizeof( face ))
			{
				printf( "%s: unexpected end of file while loading faces\n", name );
				break;
			}

			faces.push_back( face );
		}

		// load verts
		fseek( file, header.vertex_ofs, SEEK_SET );
		for( i = 0; i < header.vertex_count; i++ )
		{
			if( fread( &vert, 1, sizeof( vert ), file ) != sizeof( vert ))
			{
				printf( "%s: unexpected end of file while loading verts\n", name );
				break;
			}

			verts.push_back( vert );
		}

		// all done
		fclose( file );
		file = NULL;
		return true;
	}

	bool save( const char *name )
	{
		csm_face_t	outface;
		csm_vertex_t	outvert;
		csm_header_t	header;
		unsigned char	dummy = 0;
		int		i;

		if(( file = fopen( name, "wb" )) == NULL )
		{
			printf( "failed to save: %s\n", name );
			return false;
		}

		memset( &header, 0, sizeof( header ));
		header.ident = IDCSMMODHEADER;
		header.version = IDCSM_VERSION;
		header.header_size = sizeof( csm_header_t );
		header.flags = flags;
		header.lmGroups = lmGroups;
		header.dtGroups = dtGroups;
		header.model_mins = mins;
		header.model_maxs = maxs;

		// reserve space for header
		fwrite( &header, 1, sizeof( header ), file );

		// write materials
		header.mat_ofs = (unsigned int)ftell( file );

		for( i = 0; i < materials.size(); i++ )
			fprintf( file, "\"%s\" ", materials[i].name );
		fwrite( &dummy, 1, sizeof( dummy ), file ); // terminator
		header.mat_size = (unsigned int)ftell( file );

		// write faces
		header.faces_ofs = (unsigned int)ftell( file );
		header.face_size = sizeof( csm_face_t );
		header.faces_count = faces.size();

		// convert memory representation back to csm format
		for( i = 0; i < faces.size(); i++ )
		{
			outface = faces[i];
			fwrite( &outface, 1, sizeof( outface ), file );
		}

		// write verts
		header.vertex_ofs = (unsigned int)ftell( file );
		header.vertex_size = sizeof( csm_vertex_t );
		header.vertex_count = verts.size();

		// convert memory representation back to csm format
		for( i = 0; i < verts.size(); i++ )
		{
			outvert = verts[i];
			fwrite( &outvert, 1, sizeof( outvert ), file );
		}

		// update header
		fseek( file, 0, SEEK_SET );
		fwrite( &header, 1, sizeof( header ), file );

		// all done
		fclose( file );
		file = NULL;

		return true;
	}
};
