/*
model_smd.h - SMD Valve File Format description
Copyright (C) 2024 Uncle Mike

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
*/

#pragma once

#include "model_csm.h"
#include <ctype.h>

class model_smd_t
{
	char	line[1024];
	int	linecount;
	xform	transform;
	int	numbones;
	FILE	*file;

	bool GetLineInput( void )
	{
		while( fgets( line, sizeof( line ), file ) != NULL ) 
		{
			linecount++;
			// skip comments
			if( line[0] == '/' && line[1] == '/' )
				continue;
			return true;
		}
		return false;
	}

	bool IsEnd( char const *pLine )
	{
		if( !strcmp( "end", pLine )) 
			return true;
		return ( pLine[3] == '\0' ) || ( pLine[3] == '\n' );
	}

	int Grab_Nodes( void )
	{
		int	index, parent;
		int	numbones = 0;
		char	name[1024];

		while( GetLineInput( ))
		{
			if( sscanf( line, "%d \"%[^\"]\" %d", &index, name, &parent ) == 3 )
			{
				numbones = max( numbones, index );
				continue;
			}
			return numbones + 1;
		}

		printf( "unexpected EOF at line %d\n", linecount );

		return -1;
	}

	bool Grab_Skeleton( void )
	{
		char	cmd[1024];
		int	index;
		vec3	pos;
		rad3	rot;

		while( GetLineInput( ))
		{
			if( sscanf( line, "%d %f %f %f %f %f %f", &index, &pos.x, &pos.y, &pos.z, &rot.x, &rot.y, &rot.z ) == 7 )
			{
				if( index == 0 )
					transform = xform( pos, rot );
			}
			else if( sscanf( line, "%s %d", cmd, &index ))
			{
				if( !strcmp( cmd, "time" )) 
				{
					// begin building skeleton
				}
				else if( !strcmp( cmd, "end" )) 
				{
					return true;
				}
			}
		}

		printf( "unexpected EOF at line %d\n", linecount );

		return false;
	}

	bool Grab_Triangles( model_t *model )
	{
		// load the base triangles
		while( 1 ) 
		{
			char	texturename[CSM_MAXNAME];
			int	i;

			if( !GetLineInput( )) 
				break;

			// check for end
			if( IsEnd( line )) 
				break;

			strncpy( texturename, line, sizeof( texturename ) - 1 );
			// strip off trailing smag if present
			for( i = strlen( texturename ) - 1; i >= 0 && !isgraph( texturename[i] ); i-- );
			texturename[i + 1] = '\0';
			StripExtension( texturename );

			// weird model problem, skip them
			if( !strlen( texturename ))
			{
				GetLineInput();
				GetLineInput();
				GetLineInput();
				continue;
			}

			unsigned short	material = model->AllocMaterial( texturename );
			unsigned int	firstvert = model->verts.size();
			csm_vertex_t	verts[3];
			csm_face_t	face;

			face.material = material;

			for( int j = 0; j < 3; j++ ) 
			{
				csm_vertex_t	*srcv = &verts[j];
				int		bone;

				// quake wants them in the reverse order
				face.index[2 - j] = firstvert + j;

				if( !GetLineInput( ))
				{ 
					printf( "error on line %d: %s", linecount, line );
					return false;
				}

				// get support for Source bone weights description
				i = sscanf( line, "%d %f %f %f %f %f %f %f %f",
				&bone, 
				&srcv->point.x, &srcv->point.y, &srcv->point.z, 
				&srcv->normal.x, &srcv->normal.y, &srcv->normal.z, 
				&face.tc[FCHAN_TC].uv[2-j].x, &face.tc[FCHAN_TC].uv[2-j].y );

				if( i < 9 ) continue; 

				if( bone < 0 || bone >= numbones ) 
					printf( "bogus bone index at line %d: %s\n", linecount, line );

				// **shift everything into identity space**
				srcv->point = transform * srcv->point + transform[3];	// transform
				srcv->normal = transform * srcv->normal;		// rotate

				// flip T-coord
				face.tc[FCHAN_TC].uv[2-j].y = 1.0f - face.tc[FCHAN_TC].uv[2-j].y;

				// smd doesn't support color
				srcv->color = csm_color_t();
			}

			// add new triangle to model
			model->verts.push_back( verts[0] );
			model->verts.push_back( verts[1] );
			model->verts.push_back( verts[2] );
			model->faces.push_back( face );
		}

		return true;
	}
public:
	model_t *load( const char *name )
	{
		char	cmd[1024];
		int	option;
		model_t	*mod;

		if(( file = fopen( name, "r" )) == NULL )
		{
			printf( "failed to read: %s\n", name );
			return NULL;
		}

		transform.load_identity();
		linecount = 0;
		numbones = 0;

		mod = new model_t;

		while( GetLineInput( ))
		{
			int	numRead = sscanf( line, "%s %d", cmd, &option );

			// blank line
			if(( numRead == EOF ) || ( numRead == 0 ))
				continue;

			if( !strcmp( cmd, "version" ))
			{
				if( option != 1 )
				{
					printf( "%s version %i should be 1\n", name, option );
					fclose( file );
					delete mod;
					file = NULL;
					return NULL;
				}
			}
			else if( !strcmp( cmd, "nodes" ))
			{
				if(( numbones = Grab_Nodes()) < 0 )
				{
					fclose( file );
					delete mod;
					file = NULL;
					return NULL;
				}
			}
			else if( !strcmp( cmd, "skeleton" ))
			{
				if( !Grab_Skeleton( ))
				{
					fclose( file );
					delete mod;
					file = NULL;
					return NULL;
				}
			}
			else if( !strcmp( cmd, "triangles" ))
			{
				if( !Grab_Triangles( mod ))
				{
					fclose( file );
					delete mod;
					file = NULL;
					return NULL;
				}
			}
			else 
			{
				printf( "unknown studio command: %s\n", line );
			}
		}

		fclose( file );
		file = NULL;
		return mod;
	}

	bool save( const char *name, model_t *mdl )
	{
		if(( file = fopen( name, "w" )) == NULL )
		{
			printf( "failed to save: %s\n", name );
			return false;
		}

		// write default header
		fprintf( file, "version 1\n" );
		fprintf( file, "nodes\n" );
		fprintf( file, "  0 \"root\"  -1\n" );
		fprintf( file, "end\n" );
		fprintf( file, "skeleton\n" );
		fprintf( file, "time 0\n" );
		fprintf( file, "  0 0.000000 0.000000 0.000000 0.000000 0.000000 0.000000\n" );
		fprintf( file, "end\n" );
		fprintf( file, "triangles\n" );

		for( size_t i = 0; i < mdl->faces.size(); i++ )
		{
			const csm_face_t *face = &mdl->faces[i];

			// print material name before each triangle
			fprintf( file, "%s\n", mdl->material( face ));
			for( int k = 2; k >= 0; k-- )
			{
				// reverse triangles order
				const csm_vertex_t *v = &mdl->verts[face->index[k]];
				fprintf( file, "  0 %f %f %f %f %f %f %f %f\n", v->point.x, v->point.y, v->point.z,
				v->normal.x, v->normal.y, v->normal.z, face->tc[FCHAN_TC].uv[k].x, 1.0f - face->tc[FCHAN_TC].uv[k].y );
			}
		}

		fprintf( file, "end\n" );
		fclose( file );
		file = NULL;

		return true;
	}
};
